@@ -1,6 +1,6 @@
 # vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
 
-# Copyright 2014-2019 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
+# Copyright 2014-2020 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
 #
 # This file is part of qutebrowser.
 #
@@ -31,6 +31,8 @@ import traceback
 import warnings
 import json
 import inspect
+import typing
+import argparse
 
 from PyQt5 import QtCore
 # Optional imports
@@ -91,7 +93,10 @@ LOG_LEVELS = {
 }
 
 
-def vdebug(self, msg, *args, **kwargs):
+def vdebug(self: logging.Logger,
+           msg: str,
+           *args: typing.Any,
+           **kwargs: typing.Any) -> None:
     """Log with a VDEBUG level.
 
     VDEBUG is used when a debug message is rather verbose, and probably of
@@ -100,7 +105,7 @@ def vdebug(self, msg, *args, **kwargs):
     """
     if self.isEnabledFor(VDEBUG_LEVEL):
         # pylint: disable=protected-access
-        self._log(VDEBUG_LEVEL, msg, args, **kwargs)
+        self._log(VDEBUG_LEVEL, msg, args, **kwargs)  # type: ignore
         # pylint: enable=protected-access
 
 
@@ -151,12 +156,12 @@ LOGGER_NAMES = [
 ]
 
 
-ram_handler = None
-console_handler = None
+ram_handler = None  # type: typing.Optional[RAMHandler]
+console_handler = None  # type: typing.Optional[logging.Handler]
 console_filter = None
 
 
-def stub(suffix=''):
+def stub(suffix: str = '') -> None:
     """Show a STUB: message for the calling function."""
     try:
         function = inspect.stack()[1][3]
@@ -169,7 +174,7 @@ def stub(suffix=''):
     misc.warning(text)
 
 
-def init_log(args):
+def init_log(args: argparse.Namespace) -> None:
     """Init loggers based on the argparse namespace passed."""
     level = args.loglevel.upper()
     try:
@@ -205,18 +210,26 @@ def init_log(args):
         # disable blocks the current level (while setHandler shows the current
         # level), so -1 to avoid blocking handled messages.
         logging.disable(numeric_level - 1)
+
+    global _log_inited, _args
+    _args = args
     root.setLevel(logging.NOTSET)
     logging.captureWarnings(True)
     _init_py_warnings()
-    QtCore.qInstallMessageHandler(qt_message_handler)
-    global _log_inited, _args
+    QtCore.qInstallMessageHandler(qt_message_handler)  # type: ignore
     _log_inited = True
-    _args = args
 
 
-def _init_py_warnings():
+@QtCore.pyqtSlot()
+def shutdown_log() -> None:
+    QtCore.qInstallMessageHandler(None)
+
+
+def _init_py_warnings() -> None:
     """Initialize Python warning handling."""
-    warnings.simplefilter('default')
+    assert _args is not None
+    warnings.simplefilter('error' if 'werror' in _args.debug_flags
+                          else 'default')
     warnings.filterwarnings('ignore', module='pdb', category=ResourceWarning)
     # This happens in many qutebrowser dependencies...
     warnings.filterwarnings('ignore', category=DeprecationWarning,
@@ -226,7 +239,7 @@ def _init_py_warnings():
 
 
 @contextlib.contextmanager
-def disable_qt_msghandler():
+def disable_qt_msghandler() -> typing.Iterator[None]:
     """Contextmanager which temporarily disables the Qt message handler."""
     old_handler = QtCore.qInstallMessageHandler(None)
     try:
@@ -236,7 +249,7 @@ def disable_qt_msghandler():
 
 
 @contextlib.contextmanager
-def ignore_py_warnings(**kwargs):
+def ignore_py_warnings(**kwargs: typing.Any) -> typing.Iterator[None]:
     """Contextmanager to temporarily disable certain Python warnings."""
     warnings.filterwarnings('ignore', **kwargs)
     yield
@@ -244,7 +257,13 @@ def ignore_py_warnings(**kwargs):
         _init_py_warnings()
 
 
-def _init_handlers(level, color, force_color, json_logging, ram_capacity):
+def _init_handlers(
+        level: int,
+        color: bool,
+        force_color: bool,
+        json_logging: bool,
+        ram_capacity: int
+) -> typing.Tuple[logging.StreamHandler, typing.Optional['RAMHandler']]:
     """Init log handlers.
 
     Args:
@@ -281,7 +300,7 @@ def _init_handlers(level, color, force_color, json_logging, ram_capacity):
     return console_handler, ram_handler
 
 
-def get_console_format(level):
+def get_console_format(level: int) -> str:
     """Get the log format the console logger should use.
 
     Args:
@@ -293,7 +312,13 @@ def get_console_format(level):
     return EXTENDED_FMT if level <= logging.DEBUG else SIMPLE_FMT
 
 
-def _init_formatters(level, color, force_color, json_logging):
+def _init_formatters(
+        level: int,
+        color: bool,
+        force_color: bool,
+        json_logging: bool
+) -> typing.Tuple[typing.Union['JSONFormatter', 'ColoredFormatter'],
+                  'ColoredFormatter', 'HTMLFormatter', bool]:
     """Init log formatters.
 
     Args:
@@ -316,8 +341,8 @@ def _init_formatters(level, color, force_color, json_logging):
         return None, ram_formatter, html_formatter, False  # type: ignore
 
     if json_logging:
-        console_formatter = JSONFormatter()
-        return console_formatter, ram_formatter, html_formatter, False
+        json_formatter = JSONFormatter()
+        return json_formatter, ram_formatter, html_formatter, False
 
     use_colorama = False
     color_supported = os.name == 'posix' or colorama
@@ -334,7 +359,7 @@ def _init_formatters(level, color, force_color, json_logging):
     return console_formatter, ram_formatter, html_formatter, use_colorama
 
 
-def change_console_formatter(level):
+def change_console_formatter(level: int) -> None:
     """Change console formatter based on level.
 
     Args:
@@ -342,14 +367,16 @@ def change_console_formatter(level):
     """
     assert console_handler is not None
 
-    use_colors = console_handler.formatter.use_colors
+    old_formatter = typing.cast(ColoredFormatter, console_handler.formatter)
     console_fmt = get_console_format(level)
     console_formatter = ColoredFormatter(console_fmt, DATEFMT, '{',
-                                         use_colors=use_colors)
+                                         use_colors=old_formatter.use_colors)
     console_handler.setFormatter(console_formatter)
 
 
-def qt_message_handler(msg_type, context, msg):
+def qt_message_handler(msg_type: QtCore.QtMsgType,
+                       context: QtCore.QMessageLogContext,
+                       msg: str) -> None:
     """Qt message handler to redirect qWarning etc. to the logging system.
 
     Args:
@@ -454,13 +481,14 @@ def qt_message_handler(msg_type, context, msg):
         level = qt_to_logging[msg_type]
 
     if context.function is None:
-        func = 'none'
+        func = 'none'  # type: ignore
     elif ':' in context.function:
         func = '"{}"'.format(context.function)
     else:
         func = context.function
 
-    if context.category is None or context.category == 'default':
+    if (context.category is None or  # type: ignore
+            context.category == 'default'):
         name = 'qt'
     else:
         name = 'qt-' + context.category
@@ -472,17 +500,19 @@ def qt_message_handler(msg_type, context, msg):
                 "    pacman -S libxkbcommon-x11")
         faulthandler.disable()
 
+    assert _args is not None
     if _args.debug:
-        stack = ''.join(traceback.format_stack())
+        stack = ''.join(traceback.format_stack())  # type: typing.Optional[str]
     else:
         stack = None
+
     record = qt.makeRecord(name, level, context.file, context.line, msg, (),
                            None, func, sinfo=stack)
     qt.handle(record)
 
 
 @contextlib.contextmanager
-def hide_qt_warning(pattern, logger='qt'):
+def hide_qt_warning(pattern: str, logger: str = 'qt') -> typing.Iterator[None]:
     """Hide Qt warnings matching the given regex."""
     log_filter = QtWarningFilter(pattern)
     logger_obj = logging.getLogger(logger)
@@ -501,11 +531,11 @@ class QtWarningFilter(logging.Filter):
         _pattern: The start of the message.
     """
 
-    def __init__(self, pattern):
+    def __init__(self, pattern: str) -> None:
         super().__init__()
         self._pattern = pattern
 
-    def filter(self, record):
+    def filter(self, record: logging.LogRecord) -> bool:
         """Determine if the specified record is to be logged."""
         do_log = not record.msg.strip().startswith(self._pattern)
         return do_log
@@ -523,12 +553,13 @@ class LogFilter(logging.Filter):
         negated: Whether names is a list of records to log or to suppress.
     """
 
-    def __init__(self, names, negate=False):
+    def __init__(self, names: typing.Optional[typing.Iterable[str]],
+                 negate: bool = False) -> None:
         super().__init__()
         self.names = names
         self.negated = negate
 
-    def filter(self, record):
+    def filter(self, record: logging.LogRecord) -> bool:
         """Determine if the specified record is to be logged."""
         if self.names is None:
             return True
@@ -556,20 +587,22 @@ class RAMHandler(logging.Handler):
         _data: A deque containing the logging records.
     """
 
-    def __init__(self, capacity):
+    def __init__(self, capacity: int) -> None:
         super().__init__()
-        self.html_formatter = None
+        self.html_formatter = None  # type: typing.Optional[HTMLFormatter]
         if capacity != -1:
-            self._data = collections.deque(maxlen=capacity)
+            self._data = collections.deque(
+                maxlen=capacity
+            )  # type: typing.MutableSequence[logging.LogRecord]
         else:
             self._data = collections.deque()
 
-    def emit(self, record):
+    def emit(self, record: logging.LogRecord) -> None:
         if record.levelno >= logging.DEBUG:
             # We don't log VDEBUG to RAM.
             self._data.append(record)
 
-    def dump_log(self, html=False, level='vdebug'):
+    def dump_log(self, html: bool = False, level: str = 'vdebug') -> str:
         """Dump the complete formatted log data as string.
 
         FIXME: We should do all the HTML formatter via jinja2.
@@ -577,7 +610,13 @@ class RAMHandler(logging.Handler):
         https://github.com/qutebrowser/qutebrowser/issues/34
         """
         minlevel = LOG_LEVELS.get(level.upper(), VDEBUG_LEVEL)
-        fmt = self.html_formatter.format if html else self.format
+
+        if html:
+            assert self.html_formatter is not None
+            fmt = self.html_formatter.format
+        else:
+            fmt = self.format
+
         self.acquire()
         try:
             lines = [fmt(record)
@@ -587,7 +626,7 @@ class RAMHandler(logging.Handler):
             self.release()
         return '\n'.join(lines)
 
-    def change_log_capacity(self, capacity):
+    def change_log_capacity(self, capacity: int) -> None:
         self._data = collections.deque(self._data, maxlen=capacity)
 
 
@@ -599,11 +638,14 @@ class ColoredFormatter(logging.Formatter):
         use_colors: Whether to do colored logging or not.
     """
 
-    def __init__(self, fmt, datefmt, style, *, use_colors):
+    def __init__(self, fmt: str,
+                 datefmt: str,
+                 style: str, *,
+                 use_colors: bool) -> None:
         super().__init__(fmt, datefmt, style)
         self.use_colors = use_colors
 
-    def format(self, record):
+    def format(self, record: logging.LogRecord) -> str:
         if self.use_colors:
             color_dict = dict(COLOR_ESCAPES)
             color_dict['reset'] = RESET_ESCAPE
@@ -626,7 +668,9 @@ class HTMLFormatter(logging.Formatter):
         _colordict: The colordict passed to the logger.
     """
 
-    def __init__(self, fmt, datefmt, log_colors):
+    def __init__(self, fmt: str,
+                 datefmt: str,
+                 log_colors: typing.Mapping[str, str]) -> None:
         """Constructor.
 
         Args:
@@ -635,22 +679,22 @@ class HTMLFormatter(logging.Formatter):
             log_colors: The colors to use for logging levels.
         """
         super().__init__(fmt, datefmt)
-        self._log_colors = log_colors
-        self._colordict = {}
+        self._log_colors = log_colors  # type: typing.Mapping[str, str]
+        self._colordict = {}  # type: typing.Mapping[str, str]
         # We could solve this nicer by using CSS, but for this simple case this
         # works.
         for color in COLORS:
             self._colordict[color] = '<font color="{}">'.format(color)
         self._colordict['reset'] = '</font>'
 
-    def format(self, record):
+    def format(self, record: logging.LogRecord) -> str:
         record_clone = copy.copy(record)
         record_clone.__dict__.update(self._colordict)
         if record_clone.levelname in self._log_colors:
             color = self._log_colors[record_clone.levelname]
-            record_clone.log_color = self._colordict[color]
+            record_clone.log_color = self._colordict[color]  # type: ignore
         else:
-            record_clone.log_color = ''
+            record_clone.log_color = ''  # type: ignore
         for field in ['msg', 'filename', 'funcName', 'levelname', 'module',
                       'name', 'pathname', 'processName', 'threadName']:
             data = str(getattr(record_clone, field))
@@ -660,7 +704,8 @@ class HTMLFormatter(logging.Formatter):
             msg += self._colordict['reset']
         return msg
 
-    def formatTime(self, record, datefmt=None):
+    def formatTime(self, record: logging.LogRecord,
+                   datefmt: str = None) -> str:
         out = super().formatTime(record, datefmt)
         return pyhtml.escape(out)
 
@@ -669,7 +714,7 @@ class JSONFormatter(logging.Formatter):
 
     """Formatter for JSON-encoded log messages."""
 
-    def format(self, record):
+    def format(self, record: logging.LogRecord) -> str:
         obj = {}
         for field in ['created', 'msecs', 'levelname', 'name', 'module',
                       'funcName', 'lineno', 'levelno']:
